/* -*- mode: c++ -*- */
#ifndef __DataBuffer_h__
#define __DataBuffer_h__

#include <memory>
#include <queue>

/**
 * A DataBlock is a chunck of memory storing data.
 * It maintains an internal reference countr and when
 * it is below 0, that is, nobody is interested in the
 * memory chunck it will deallocate it.
 */
class DataBlock
{
      /// size of the buffer
      unsigned int _size;
      /// pointer to the buffer
      char *_data;

      /// used internally to implement copy, construction and assignment
      void cpy(const DataBlock &e);

      /// The event allocator
      static std::allocator<DataBlock> buffer_allocator;
      /// The data allocator
      static std::allocator<char> data_allocator;

   public:
      /// Constructor
      DataBlock(int s, const char *d=0);
      /// Constructor
      DataBlock(const DataBlock &e) :
         _size(0), _data(0)
      {
         cpy(e);
      }
      /// Asignment operator
      DataBlock &operator=(const DataBlock &e)
      {
         if ( &e != this)
            cpy(e);
         return *this;
      }

      /// Destructor
      ~DataBlock();

      /// Returns the size of the data
      unsigned int size() const
      {
         return _size;
      }
      /// A pointr to the data buffer
      const char *data() const
      {
         return _data;
      }

      /// new is overloaded
      inline void* operator new(size_t); /// overloaded new operator
      /// delete is overloaded
      inline void operator delete(void*); /// overloaded delete operator
};

// overloaded new operator
inline void* DataBlock::operator new(size_t)
{
    return buffer_allocator.allocate(1);
}

// implementation of overloaded delete operator
inline void DataBlock::operator delete(void* aDataBlock)
{
    buffer_allocator.deallocate((DataBlock *)aDataBlock, 1);

}

class DataBuffer : public std::queue<DataBlock *>
{
   private:
      unsigned int _mxsize;
   public:
      DataBuffer(int sz = 50);
      ~DataBuffer();

      void push_block(DataBlock *b);
      DataBlock *get_block();

      int items_left() const { return size(); }
      void max_size(unsigned int sz)
      {
         _mxsize = sz;
      }
      unsigned int max_size() const
      {
         return _mxsize;
      }

      bool full() const
      {
         return size() > _mxsize;
      }

      void flush();
};

#endif
