//#define DEBUG 1
#include <iostream>
#include <cstdlib>
#include <cstring>
#include <ctime>
#include <unistd.h>
#include <glib.h>
#ifdef G_OS_UNIX
#include <sys/ioctl.h>
#include <poll.h>
#include <termios.h>
#endif
#include <cerrno>
#include "USBpatch.h"
#include "Fifo.h"
#include "utils.h"

#ifndef FIONREAD
#define FIONREAD TIOCINQ
#endif

#include "USBport.h"
#ifdef G_OS_WIN32
#include "USBserialWnds.h"
#include "mingw.h"
#endif


Fifo::Fifo(unsigned int size) :
    cond(0), cond_mutex(0), data_mutex(0), thread(0),
#ifdef HAVE_GIO
    sckt(0), cancellable(0),
#endif
    fd(0), in_loop(false),
    _end_poll(false),bfsiz(size), buff(0), _P(0)
{
    if (!g_thread_supported ())
        g_thread_init(NULL);

    buff = reinterpret_cast<char *>(realloc(buff, size * sizeof(char)));
    rdptr = wrptr = buff;
    end = buff + bfsiz;

    cond = g_cond_new();
    cond_mutex = g_mutex_new();
    data_mutex = g_mutex_new();

}

Fifo::~Fifo()
{
    detach();

    free(buff);
    g_cond_free(cond);
    g_mutex_free(cond_mutex);
    g_mutex_free(data_mutex);
}

void Fifo::endpoll(bool x)
{
    _end_poll=x;
#ifdef G_OS_WIN32
    if (_P)
    {
        USBserialWnds *_W = dynamic_cast<USBserialWnds *>(_P);
        _W->stop_thread();
    }
#endif
}


void Fifo::resize(unsigned int sz)
{
    g_mutex_lock(data_mutex);
    if (sz>bfsiz)
    {
        grow(sz-bfsiz);
    }
    g_mutex_unlock(data_mutex);
}

bool Fifo::wait(double tmout)
{
    PDEBUG("wait with tmout=%f\n",tmout);
    if (tmout>0.)
    {
        GTimeVal tim;
        g_get_current_time(&tim);
        g_time_val_add(&tim, (glong)(tmout*1000000));
        g_mutex_lock (cond_mutex);
        while (empty())
        {
            g_cond_timed_wait(cond, cond_mutex, &tim);
            break;
        }
        g_mutex_unlock(cond_mutex);
    }
    else
    {
        g_mutex_lock (cond_mutex);
        while (empty() && !endpoll())
        {
            g_cond_wait (cond, cond_mutex);
        }
        g_mutex_unlock (cond_mutex);
    }
    PDEBUG("...stop waiting\n");
    return empty() ? false : true;
}

void Fifo::notify()
{
    PDEBUG("notify\n");
    g_mutex_lock (cond_mutex);
    g_cond_signal(cond);
    g_mutex_unlock (cond_mutex);
}

int Fifo::flush()
{
    GError *err=0;
    gsize n_in=0;
    PDEBUG("...trying to lock\n");
    g_mutex_lock(data_mutex);
    PDEBUG("...lock\n");

    reset();

    int rc, nread=0;

#ifdef G_OS_WIN32
    /**
     * TODO: how to handle non sockets in windows...
     */
    if (sckt)
    {
        int rc;
        unsigned long uread;
        if ( (rc = ioctlsocket(g_socket_get_fd(sckt),FIONREAD, &uread) ) )
        {
            g_printerr("Error calling ioctlsocket\n");
        }
        else
            nread = uread;
    }
//    do
//    {
//        rc = from_source(buff, 1);
//    } while (rc>0);
#else
    ioctl(fd, FIONREAD, &nread);
#endif
    if (nread > 0)
    {
        char *buff = new char[nread];
        g_io_channel_read_chars(chan, buff, nread, &n_in, &err);
        delete [] buff;
    }
    g_mutex_unlock(data_mutex);
    PDEBUG("...unlock\n");
    return 0;
}


int Fifo::from_file(void *buf, int size)
{
    return read_from_file(fd, buf, size);
}
int Fifo::from_source(void *buf, int size)
{
    GError *err=0;
    gsize sz=0;
    GIOStatus status;
    status = g_io_channel_read_chars(chan, (gchar *)buf, size, &sz, &err);
    if (err)
    {
        g_printerr("Fifo: Error reading from channel: %s\n", err->message);
        g_error_free(err);
    }
    return sz;
}


int Fifo::read(char *data, int count)
{
    int rc = 0;

    PDEBUG("read: size=%d\n", count);
    wait();

    PDEBUG("...trying to lock\n");
    g_mutex_lock(data_mutex);
    PDEBUG("...lock\n");
    if (wrptr > rdptr)
    {
        // write pointer has not wrapped yet
        if (count < wrptr - rdptr)
        { // all the data is there
            memcpy(data, rdptr, count);
            rc = count;
            rdptr += count;
        }
        else
        { // some data is missing
            rc = wrptr - rdptr;
            memcpy(data, rdptr, rc);
            rdptr += rc;
        }
    }
    else
    { // we may have to wrap
        int left = space_left();
        if (count <= left)
        { // all the data is there
            if (count < end - rdptr)
            { // no need to wrap
                memcpy(data, rdptr, count);
                rdptr += count;
            }
            else
            { // we do need to wrap
                rc = end - rdptr;
                memcpy(data, rdptr, rc);
                rdptr = buff;
                memcpy(data + rc, rdptr, count - rc);
                rdptr += count - rc;
            }
            rc = count;
        }
        else
        { // only some data is there
            int sz;
            rc = end - rdptr;
            memcpy(data, rdptr, rc);
            rdptr = buff;
            sz = wrptr - rdptr;
            memcpy(data + rc, rdptr, sz);
            rdptr += sz;
            rc += sz;
        }
    }

    g_mutex_unlock(data_mutex);
    PDEBUG("...unlock\n");
    PDEBUG("...rc=%d\n",rc);
    return rc;
}

void Fifo::grow(int count)
{
    PDEBUG("Have to grow: %d\n",count);
    int nsz = bfsiz + count;
    int nw = (wrptr - buff);
    int nr = (rdptr - buff);
    int nt = data_available();
    if (wrptr > rdptr)
    {
        buff = reinterpret_cast<char *>(realloc(buff, nsz));
        if (buff==0)
        {
            err_msg(_("Ran out of memory"));
            exit(1);
        }
        bfsiz = nsz;
        end = buff + bfsiz;
        wrptr = buff + nw;
        rdptr = buff + nr;
    }
    else
    {
        char *nbuff = reinterpret_cast<char *>(realloc(0, nsz));
        if (nbuff==0)
        {
            err_msg(_("Ran out of memory"));
            exit(1);
        }
//        char *ptr, *qtr=nbuff;
//        for (ptr=rdptr; ptr<end;)
//        {
//            *qtr++ = *ptr++;
//        }
        memcpy(nbuff, rdptr, (end-rdptr));

//        for (ptr=buff; ptr<wrptr;)
//        {
//            *qtr++ = *ptr++;
//        }
        memcpy(nbuff+(end-rdptr), buff, (wrptr-buff));
        free(buff);
        buff = nbuff;
        bfsiz = nsz;
        end = buff + bfsiz;
        rdptr = buff;
        wrptr = buff+nt;
    }
}

int Fifo::write(char *data, int count, bool force)
{
    int rc = 0, left, sz;

    g_mutex_lock(data_mutex);
    left = space_left();
    if (left <= count)
    {
        if ( !force)
        {
            g_mutex_unlock(data_mutex);
            return rc;
        }

        // we need to enlarge the buffer
        grow(count);
    }

    if (wrptr >= rdptr)
    { // The write pointer has not wrapped yet
        if (count < end - wrptr)
        { // no need to wrap
            memcpy(wrptr, data, count);
            wrptr += count;
        }
        else
        { // we will have to wrap
            sz = end - wrptr;
            memcpy(wrptr, data, sz);
            wrptr = buff;
            memcpy(wrptr, data + sz, count - sz);
            wrptr += count - sz;
        }
        rc = count;
    }
    else
    {
        memcpy(wrptr, data, count);
        wrptr += count;
        rc = count;
    }
    notify();
    g_mutex_unlock(data_mutex);
    return rc;
}

int Fifo::attach(int fildes)
{
    PDEBUG("attach\n");
    GError *gerror=0;
#ifdef G_OS_WIN32
    chan = g_io_channel_win32_new_fd( fildes );
#else
    chan = g_io_channel_unix_new( fildes );
#endif
    g_io_channel_set_encoding(chan, NULL, &gerror);
    if (gerror)
    {
    	err_msg(_("Could not create the thread: %s"), gerror->message);
    	g_error_free(gerror);
    	gerror = 0;
    }
#if HAVE_GIO
    if (cancellable)
    {
        g_object_unref(cancellable);
        cancellable = 0;
    }
#endif
    fd = fildes;
    in_loop = false;
    PDEBUG(_("...starting the thread\n"));
    thread = g_thread_create((GThreadFunc)start_loop,this, TRUE, &gerror);
    if (!thread)
    {
    	if ( gerror ) {
    		err_msg(_("Could not create the thread: %s"), gerror->message);
    		g_error_free(gerror);
    		gerror = 0;
    	}
    	else
    		err_msg(_("Could not create the thread"));
    }
    else
    {
        while (!in_loop)
        {
            g_usleep(1);
        }
        PDEBUG("...thread up and running\n");
    }

    return 0;
}

/*
 * TODO: the 2 attach should be refactored... have too much in common
 */
#ifdef HAVE_GIO
int Fifo::attach(GSocket *S, GIOChannel *C)
{
    GError *gerror=0;
    sckt = S;
    chan = C;
    cancellable = g_cancellable_new();
    fd = 0;
#ifdef G_OS_UNIX
    fd = g_socket_get_fd(S);
#endif

    /*
     * If we do it like this we are not setting the in_loop variable
     */
    endpoll(false);
#ifdef USE_G_LOOP
    in_loop = true;
    g_io_add_watch(chan, (GIOCondition)(G_IO_IN|G_IO_HUP), (GIOFunc)data_from_channel, this);
#else
    in_loop = false;
    PDEBUG("...starting the thread\n");
    thread = g_thread_create((GThreadFunc)start_loop,this, TRUE, &gerror);
    if ( gerror )
    {
        err_msg(_("Could not create the thread: %s"), gerror->message);
        g_error_free( gerror );
    }
    else
    {
        while (!in_loop)
        {
            g_usleep(1);
        }
        PDEBUG(_("...thread up and running\n"));
    }

#endif
    return 0;

}
#endif

#ifdef G_OS_WIN32
HANDLE get_handle_from_fd(int fd);

int Fifo::attach(USBport *port)
{
	GError *gerror = 0;
    if (cancellable)
    {
        g_object_unref(cancellable);
        cancellable = 0;
    }
    _P = port;
    fd = port->fildes();
    PDEBUG("...starting the thread\n");

    if (!SetCommMask(get_handle_from_fd(fd), EV_RXCHAR|EV_TXEMPTY) )
    {
        char lastError[1024];
        FormatMessage(
                FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
                NULL,
                GetLastError(),
                MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                lastError,
                1024,
                NULL);
        err_msg("%s", lastError);
        return 1;
    }

    in_loop = true;
    thread = g_thread_create((GThreadFunc)start_win_loop,this, TRUE, &gerror);
    if (!thread)
    {
        in_loop = false;
        if ( gerror ) {
            err_msg(_("Could not create the thread: %s"), gerror->message);
            g_error_free(gerror);
            gerror = 0;
        }
        else
            err_msg(_("Could not create the thread"));
    }


    return 0;
}


gpointer Fifo::start_win_loop(Fifo *F)
{
    USBserialWnds *_W = dynamic_cast<USBserialWnds *>(F->_P);
    _W->wait_for_port_data(F);
    return 0;
}

#endif

void Fifo::detach()
{
#ifdef G_OS_WIN32
	if (_P)
	{
		USBserialWnds *_W = dynamic_cast<USBserialWnds *>(_P);
		_W->stop_thread();
	}
#endif
    if (!thread)
        return;

    endpoll(true);

#if HAVE_GIO
    if (cancellable)
        g_cancellable_cancel(cancellable);
#endif

    g_thread_join(thread);
}

void Fifo::read_from_channel()
{
    int rc, sz, left;

    int nread=256;
#ifdef G_OS_WIN32
    if ( sckt )
    {
        int rc;
        unsigned long uread;
        if ( (rc = ioctlsocket(g_socket_get_fd(sckt),FIONREAD, &uread) ) )
        {
            g_printerr("Error calling ioctlsocket\n");
        }
        else
            nread = uread;
    }
#else
    if ( fd > 0 )
    {
        ioctl(fd, FIONREAD, &nread);
        if (nread<=0)
            return;
    }
#endif

    PDEBUG("reading %d from file\n",nread);

    PDEBUG("...trying to lock\n");
    g_mutex_lock(data_mutex);
    PDEBUG("...lock\n");
    left = space_left();
    if (left<=nread)
    {
        grow(nread);
    }

    if (wrptr >= rdptr)
    { // The write pointer has not wrapped yet
        if (nread < end-wrptr)
        {
            // no need to wrap yet
            rc = from_source(wrptr, nread);
            if (rc>0)
            {
                wrptr += rc;
            }
        }
        else
        {
            // we will have to wrap
            sz = end-wrptr;
            rc = from_source(wrptr, sz);
            wrptr = buff;
            rc = from_source(wrptr, nread-sz);
            wrptr += rc; // (nread-sz);
        }
    }
    else
    {
        rc = from_source(wrptr, nread);
        wrptr += rc;
    }
    PDEBUG("...going to notify...");
    notify();
    PDEBUG("...unloking\n");
    g_mutex_unlock(data_mutex);
    PDEBUG("...OK\n");
}

gboolean Fifo::data_from_channel(GIOChannel *source, GIOCondition condition, Fifo *fifo)
{
    if (condition != G_IO_IN)
    {
        err_msg(_("Error reading from channel. Condition: %d"), condition);
        if ( (condition & G_IO_HUP) || (condition & G_IO_ERR))
        {
            err_msg(_("...quiting the loop"));
            return FALSE;
        }
    }
    PDEBUG(_("Fifo::data_from_channel\n"));
    fifo->read_from_channel();
    return fifo->endpoll() ? false : true;
}

#if GLIB_CHECK_VERSION(2, 20, 0)
gpointer Fifo::start_loop(Fifo *fifo)
{
#ifdef G_OS_WIN32
    /*
     * NOTE: In windows we can only have a channel...
     */
    GPollFD gfd;
    g_io_channel_win32_make_pollfd (fifo->chan,
                                    (GIOCondition)(G_IO_IN | G_IO_ERR | G_IO_HUP),
                                    &gfd);
#else
    GPollFD gfd = { 0, (GIOCondition)(G_IO_IN | G_IO_ERR | G_IO_HUP), 0 };
    if (fifo->fd>0)
    {
        gfd.fd = fifo->fd;
    }
    else if (fifo->chan)
    {
        gfd.fd = g_io_channel_unix_get_fd(fifo->chan);
    }
#ifdef HAVE_GIO
    else if ( fifo->sckt )
    {
        gfd.fd = g_socket_get_fd(fifo->sckt);
    }
#endif
#endif

    fifo->endpoll(false);
    while (true)
    {
        fifo->in_loop = true;
        int iok = g_poll(&gfd, 1, 1000);
        if (iok==1)
        {
            if ( gfd.revents & G_IO_IN )
            {
                fifo->read_from_channel();
            }
            {
                if ( (gfd.revents & G_IO_HUP) || (gfd.revents & G_IO_ERR) )
                {
                    err_msg(_("POLLHUP: Cable disconnected?\n"));
                    fifo->endpoll(true);
                    g_cond_broadcast( fifo->cond );
                    break;
                }
            }
        }
        else
        {
            if (iok<0)
            {
                if (errno==EINTR)
                {
                    if (fifo->endpoll())
                        return NULL;
                }
                else
                {
                    err_msg(_("error polling data"));
                    return NULL;
                }
            }
            else if (fifo->endpoll())
            {
                return NULL;
            }
        }

    }
    return NULL;
}
#else
gpointer Fifo::start_loop(Fifo *fifo)
{
    struct pollfd PS = { fifo->fd, POLLIN | POLLPRI | POLLERR | POLLHUP | POLLNVAL, 0 };

    fifo->endpoll(false);
    while (true)
    {
        fifo->in_loop = true;
        int iok = poll(&PS, 1, 1000);
        if (iok==1)
        {
            if ( PS.revents & POLLIN )
                fifo->read_from_channel();
            {
                if ( PS.revents & POLLHUP )
                {
                    err_msg("POLLHUP: Cable disconnected?\n");
                }
                if ( PS.revents & POLLERR )
                {
                    err_msg("POLLERR\n");
                }
                if ( PS.revents & POLLNVAL )
                {
                    err_msg("POLLNVAL\n");
                }
            }
        }
        else
        {
            if (iok<0)
            {
                if (errno==EINTR)
                {
                    if (fifo->endpoll())
                        return NULL;
                }
                else
                {
                    err_msg("error polling data\n");
                    return NULL;
                }
            }
            else if (fifo->endpoll())
            {
                return NULL;
            }
        }

    }
    return NULL;
}

#endif
